@@ -55,7 +55,7 @@ G_DEFINE_TYPE (GstDeinterlaceMethod, gst_deinterlace_method, GST_TYPE_OBJECT);
 static void
 gst_deinterlace_method_class_init (GstDeinterlaceMethodClass * klass)
 {
-  klass->available = TRUE;
+
 }
 
 static void
@@ -89,6 +89,166 @@ gst_deinterlace_method_get_latency (GstDeinterlaceMethod * self)
   return klass->latency;
 }
 
+
+G_DEFINE_TYPE (GstDeinterlaceSimpleMethod, gst_deinterlace_simple_method,
+    GST_TYPE_DEINTERLACE_METHOD);
+
+static void
+gst_deinterlace_simple_method_interpolate_scanline (GstDeinterlaceMethod * self,
+    GstDeinterlace2 * parent, guint8 * out,
+    GstDeinterlaceScanlineData * scanlines, gint width)
+{
+  oil_memcpy (out, scanlines->m1, parent->line_length);
+}
+
+static void
+gst_deinterlace_simple_method_copy_scanline (GstDeinterlaceMethod * self,
+    GstDeinterlace2 * parent, guint8 * out,
+    GstDeinterlaceScanlineData * scanlines, gint width)
+{
+  oil_memcpy (out, scanlines->m0, parent->line_length);
+}
+
+static void
+gst_deinterlace_simple_method_deinterlace_frame (GstDeinterlaceMethod * self,
+    GstDeinterlace2 * parent)
+{
+  GstDeinterlaceSimpleMethodClass *dsm_class =
+      GST_DEINTERLACE_SIMPLE_METHOD_GET_CLASS (self);
+  GstDeinterlaceMethodClass *dm_class = GST_DEINTERLACE_METHOD_GET_CLASS (self);
+  GstDeinterlaceScanlineData scanlines;
+  guint8 *out = GST_BUFFER_DATA (parent->out_buf);
+  guint8 *field0 = NULL, *field1 = NULL, *field2 = NULL, *field3 = NULL;
+  gint cur_field_idx = parent->history_count - dm_class->fields_required;
+  guint cur_field_flags = parent->field_history[cur_field_idx].flags;
+  gint line;
+
+  field0 = GST_BUFFER_DATA (parent->field_history[cur_field_idx].buf);
+
+  g_assert (dm_class->fields_required <= 4);
+
+  if (dm_class->fields_required >= 2)
+    field1 = GST_BUFFER_DATA (parent->field_history[cur_field_idx + 1].buf);
+  if (dm_class->fields_required >= 3)
+    field2 = GST_BUFFER_DATA (parent->field_history[cur_field_idx + 2].buf);
+  if (dm_class->fields_required >= 4)
+    field3 = GST_BUFFER_DATA (parent->field_history[cur_field_idx + 3].buf);
+
+
+  if (cur_field_flags == PICTURE_INTERLACED_BOTTOM) {
+    /* double the first scanline of the bottom field */
+    oil_memcpy (out, field0, parent->line_length);
+    out += parent->output_stride;
+  }
+
+  oil_memcpy (out, field0, parent->line_length);
+  out += parent->output_stride;
+
+  for (line = 2; line <= parent->field_height; line++) {
+
+    memset (&scanlines, 0, sizeof (scanlines));
+    scanlines.bottom_field = (cur_field_flags == PICTURE_INTERLACED_BOTTOM);
+
+    /* interp. scanline */
+    scanlines.t0 = field0;
+    scanlines.b0 = field0 + parent->field_stride;
+
+    if (field1 != NULL) {
+      scanlines.tt1 = field1;
+      scanlines.m1 = field1 + parent->field_stride;
+      scanlines.bb1 = field1 + parent->field_stride * 2;
+      field1 += parent->field_stride;
+    }
+
+    if (field2 != NULL) {
+      scanlines.t2 = field2;
+      scanlines.b2 = field2 + parent->field_stride;
+    }
+
+    if (field3 != NULL) {
+      scanlines.tt3 = field3;
+      scanlines.m3 = field3 + parent->field_stride;
+      scanlines.bb3 = field3 + parent->field_stride * 2;
+      field3 += parent->field_stride;
+    }
+
+    /* set valid data for corner cases */
+    if (line == 2) {
+      scanlines.tt1 = scanlines.bb1;
+      scanlines.tt3 = scanlines.bb3;
+    } else if (line == parent->field_height) {
+      scanlines.bb1 = scanlines.tt1;
+      scanlines.bb3 = scanlines.tt3;
+    }
+
+    dsm_class->interpolate_scanline (self, parent, out, &scanlines,
+        parent->frame_width);
+    out += parent->output_stride;
+
+    memset (&scanlines, 0, sizeof (scanlines));
+    scanlines.bottom_field = (cur_field_flags == PICTURE_INTERLACED_BOTTOM);
+
+    /* copy a scanline */
+    scanlines.tt0 = field0;
+    scanlines.m0 = field0 + parent->field_stride;
+    scanlines.bb0 = field0 + parent->field_stride * 2;
+    field0 += parent->field_stride;
+
+    if (field1 != NULL) {
+      scanlines.t1 = field1;
+      scanlines.b1 = field1 + parent->field_stride;
+    }
+
+    if (field2 != NULL) {
+      scanlines.tt2 = field2;
+      scanlines.m2 = field2 + parent->field_stride;
+      scanlines.bb2 = field2 + parent->field_stride * 2;
+      field2 += parent->field_stride;
+    }
+
+    if (field3 != NULL) {
+      scanlines.t3 = field3;
+      scanlines.b3 = field3 + parent->field_stride;
+    }
+
+    /* set valid data for corner cases */
+    if (line == parent->field_height) {
+      scanlines.bb0 = scanlines.tt0;
+      scanlines.b1 = scanlines.t1;
+      scanlines.bb2 = scanlines.tt2;
+      scanlines.b3 = scanlines.t3;
+    }
+
+    dsm_class->copy_scanline (self, parent, out, &scanlines,
+        parent->frame_width);
+    out += parent->output_stride;
+  }
+
+  if (cur_field_flags == PICTURE_INTERLACED_TOP) {
+    /* double the last scanline of the top field */
+    oil_memcpy (out, field0, parent->line_length);
+  }
+}
+
+static void
+gst_deinterlace_simple_method_class_init (GstDeinterlaceSimpleMethodClass *
+    klass)
+{
+  GstDeinterlaceMethodClass *dm_class = (GstDeinterlaceMethodClass *) klass;
+
+  dm_class->deinterlace_frame = gst_deinterlace_simple_method_deinterlace_frame;
+  dm_class->fields_required = 2;
+
+  klass->interpolate_scanline =
+      gst_deinterlace_simple_method_interpolate_scanline;
+  klass->copy_scanline = gst_deinterlace_simple_method_copy_scanline;
+}
+
+static void
+gst_deinterlace_simple_method_init (GstDeinterlaceSimpleMethod * self)
+{
+}
+
 #define GST_TYPE_DEINTERLACE2_METHODS (gst_deinterlace2_methods_get_type ())
 static GType
 gst_deinterlace2_methods_get_type (void)
@@ -96,12 +256,18 @@ gst_deinterlace2_methods_get_type (void)
   static GType deinterlace2_methods_type = 0;
 
   static const GEnumValue methods_types[] = {
-#if 0
-    {GST_DEINTERLACE2_TOMSMOCOMP, "Toms Motion Compensation", "tomsmocomp"},
-#endif
-    {GST_DEINTERLACE2_GREEDY_H, "Greedy High Motion", "greedyh"},
-    {GST_DEINTERLACE2_GREEDY_L, "Greedy Low Motion", "greedyl"},
-    {GST_DEINTERLACE2_VFIR, "Vertical Blur", "vfir"},
+    {GST_DEINTERLACE2_TOMSMOCOMP, "Motion Adaptive: Motion Search",
+        "tomsmocomp"},
+    {GST_DEINTERLACE2_GREEDY_H, "Motion Adaptive: Advanced Detection",
+        "greedyh"},
+    {GST_DEINTERLACE2_GREEDY_L, "Motion Adaptive: Simple Detection", "greedyl"},
+    {GST_DEINTERLACE2_VFIR, "Blur Vertical", "vfir"},
+    {GST_DEINTERLACE2_LINEAR, "Television: Full resolution", "linear"},
+    {GST_DEINTERLACE2_LINEAR_BLEND, "Blur: Temporal", "linearblend"},
+    {GST_DEINTERLACE2_SCALER_BOB, "Double lines", "scalerbob"},
+    {GST_DEINTERLACE2_WEAVE, "Weave", "weave"},
+    {GST_DEINTERLACE2_WEAVE_TFF, "Progressive: Top Field First", "weavetff"},
+    {GST_DEINTERLACE2_WEAVE_BFF, "Progressive: Bottom Field First", "weavebff"},
     {0, NULL, NULL},
   };
 
@@ -171,6 +337,7 @@ static void gst_deinterlace2_set_property (GObject * self, guint prop_id,
 static void gst_deinterlace2_get_property (GObject * self, guint prop_id,
     GValue * value, GParamSpec * pspec);
 
+static GstCaps *gst_deinterlace2_getcaps (GstPad * pad);
 static gboolean gst_deinterlace2_setcaps (GstPad * pad, GstCaps * caps);
 static gboolean gst_deinterlace2_sink_event (GstPad * pad, GstEvent * event);
 static GstFlowReturn gst_deinterlace2_chain (GstPad * pad, GstBuffer * buffer);
@@ -215,11 +382,9 @@ gst_deinterlace2_set_method (GstDeinterlace2 * self,
   }
 
   switch (method) {
-#if 0
     case GST_DEINTERLACE2_TOMSMOCOMP:
       self->method = g_object_new (GST_TYPE_DEINTERLACE_TOMSMOCOMP, NULL);
       break;
-#endif
     case GST_DEINTERLACE2_GREEDY_H:
       self->method = g_object_new (GST_TYPE_DEINTERLACE_GREEDY_H, NULL);
       break;
@@ -229,6 +394,24 @@ gst_deinterlace2_set_method (GstDeinterlace2 * self,
     case GST_DEINTERLACE2_VFIR:
       self->method = g_object_new (GST_TYPE_DEINTERLACE_VFIR, NULL);
       break;
+    case GST_DEINTERLACE2_LINEAR:
+      self->method = g_object_new (GST_TYPE_DEINTERLACE_LINEAR, NULL);
+      break;
+    case GST_DEINTERLACE2_LINEAR_BLEND:
+      self->method = g_object_new (GST_TYPE_DEINTERLACE_LINEAR_BLEND, NULL);
+      break;
+    case GST_DEINTERLACE2_SCALER_BOB:
+      self->method = g_object_new (GST_TYPE_DEINTERLACE_SCALER_BOB, NULL);
+      break;
+    case GST_DEINTERLACE2_WEAVE:
+      self->method = g_object_new (GST_TYPE_DEINTERLACE_WEAVE, NULL);
+      break;
+    case GST_DEINTERLACE2_WEAVE_TFF:
+      self->method = g_object_new (GST_TYPE_DEINTERLACE_WEAVE_TFF, NULL);
+      break;
+    case GST_DEINTERLACE2_WEAVE_BFF:
+      self->method = g_object_new (GST_TYPE_DEINTERLACE_WEAVE_BFF, NULL);
+      break;
     default:
       GST_WARNING ("Invalid Deinterlacer Method");
       return;
@@ -275,7 +458,8 @@ gst_deinterlace2_class_init (GstDeinterlace2Class * klass)
           "Method",
           "Deinterlace Method",
           GST_TYPE_DEINTERLACE2_METHODS,
-          GST_DEINTERLACE2_GREEDY_H, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)
+          GST_DEINTERLACE2_TOMSMOCOMP,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)
       );
 
   g_object_class_install_property (gobject_class, ARG_FIELDS,
@@ -337,7 +521,7 @@ gst_deinterlace2_init (GstDeinterlace2 * self, GstDeinterlace2Class * klass)
   gst_pad_set_setcaps_function (self->sinkpad,
       GST_DEBUG_FUNCPTR (gst_deinterlace2_setcaps));
   gst_pad_set_getcaps_function (self->sinkpad,
-      GST_DEBUG_FUNCPTR (gst_pad_proxy_getcaps));
+      GST_DEBUG_FUNCPTR (gst_deinterlace2_getcaps));
   gst_element_add_pad (GST_ELEMENT (self), self->sinkpad);
 
   self->srcpad = gst_pad_new_from_static_template (&src_templ, "src");
@@ -350,7 +534,7 @@ gst_deinterlace2_init (GstDeinterlace2 * self, GstDeinterlace2Class * klass)
   gst_pad_set_setcaps_function (self->srcpad,
       GST_DEBUG_FUNCPTR (gst_deinterlace2_setcaps));
   gst_pad_set_getcaps_function (self->srcpad,
-      GST_DEBUG_FUNCPTR (gst_pad_proxy_getcaps));
+      GST_DEBUG_FUNCPTR (gst_deinterlace2_getcaps));
   gst_element_add_pad (GST_ELEMENT (self), self->srcpad);
 
   gst_element_no_more_pads (GST_ELEMENT (self));
@@ -594,9 +778,14 @@ gst_deinterlace2_chain (GstPad * pad, GstBuffer * buf)
       /* do magic calculus */
       gst_deinterlace_method_deinterlace_frame (self->method, self);
 
-      buf = gst_deinterlace2_pop_history (self);
+      g_assert (self->history_count - 1 -
+          gst_deinterlace_method_get_latency (self->method) >= 0);
+      buf =
+          self->field_history[self->history_count - 1 -
+          gst_deinterlace_method_get_latency (self->method)].buf;
       timestamp = GST_BUFFER_TIMESTAMP (buf);
-      gst_buffer_unref (buf);
+
+      gst_buffer_unref (gst_deinterlace2_pop_history (self));
 
       GST_BUFFER_TIMESTAMP (self->out_buf) = timestamp;
       if (self->fields == GST_DEINTERLACE2_ALL)
@@ -613,8 +802,7 @@ gst_deinterlace2_chain (GstPad * pad, GstBuffer * buf)
     else if (self->field_history[cur_field_idx].flags ==
         PICTURE_INTERLACED_TOP && self->fields == GST_DEINTERLACE2_BF) {
       GST_DEBUG ("Removing unused top field");
-      buf = gst_deinterlace2_pop_history (self);
-      gst_buffer_unref (buf);
+      gst_buffer_unref (gst_deinterlace2_pop_history (self));
     }
 
     cur_field_idx = self->history_count - fields_required;
@@ -637,9 +825,14 @@ gst_deinterlace2_chain (GstPad * pad, GstBuffer * buf)
       /* do magic calculus */
       gst_deinterlace_method_deinterlace_frame (self->method, self);
 
-      buf = gst_deinterlace2_pop_history (self);
+      g_assert (self->history_count - 1 -
+          gst_deinterlace_method_get_latency (self->method) >= 0);
+      buf =
+          self->field_history[self->history_count - 1 -
+          gst_deinterlace_method_get_latency (self->method)].buf;
       timestamp = GST_BUFFER_TIMESTAMP (buf);
-      gst_buffer_unref (buf);
+
+      gst_buffer_unref (gst_deinterlace2_pop_history (self));
 
       GST_BUFFER_TIMESTAMP (self->out_buf) = timestamp;
       if (self->fields == GST_DEINTERLACE2_ALL)
@@ -657,8 +850,7 @@ gst_deinterlace2_chain (GstPad * pad, GstBuffer * buf)
     else if (self->field_history[cur_field_idx].flags ==
         PICTURE_INTERLACED_BOTTOM && self->fields == GST_DEINTERLACE2_TF) {
       GST_DEBUG ("Removing unused bottom field");
-      buf = gst_deinterlace2_pop_history (self);
-      gst_buffer_unref (buf);
+      gst_buffer_unref (gst_deinterlace2_pop_history (self));
     }
   }
 
@@ -667,6 +859,197 @@ gst_deinterlace2_chain (GstPad * pad, GstBuffer * buf)
   return ret;
 }
 
+static gint
+gst_greatest_common_divisor (gint a, gint b)
+{
+  while (b != 0) {
+    int temp = a;
+
+    a = b;
+    b = temp % b;
+  }
+
+  return ABS (a);
+}
+
+static gboolean
+gst_fraction_double (gint * n_out, gint * d_out, gboolean half)
+{
+  gint n, d, gcd;
+
+  n = *n_out;
+  d = *d_out;
+
+  if (d == 0)
+    return FALSE;
+
+  if (n == 0 || (n == G_MAXINT && d == 1))
+    return TRUE;
+
+  gcd = gst_greatest_common_divisor (n, d);
+  n /= gcd;
+  d /= gcd;
+
+  if (!half) {
+    if (G_MAXINT / 2 >= ABS (n)) {
+      n *= 2;
+    } else if (d >= 2) {
+      d /= 2;
+    } else {
+      return FALSE;
+    }
+  } else {
+    if (G_MAXINT / 2 >= ABS (d)) {
+      d *= 2;
+    } else if (n >= 2) {
+      n /= 2;
+    } else {
+      return FALSE;
+    }
+  }
+
+  *n_out = n;
+  *d_out = d;
+
+  return TRUE;
+}
+
+static GstCaps *
+gst_deinterlace2_getcaps (GstPad * pad)
+{
+  GstCaps *ret;
+  GstDeinterlace2 *self = GST_DEINTERLACE2 (gst_pad_get_parent (pad));
+  GstPad *otherpad;
+  gint len;
+  const GstCaps *ourcaps;
+  GstCaps *peercaps;
+
+  GST_OBJECT_LOCK (self);
+
+  otherpad = (pad == self->srcpad) ? self->sinkpad : self->srcpad;
+
+  ourcaps = gst_pad_get_pad_template_caps (pad);
+  peercaps = gst_pad_peer_get_caps (otherpad);
+
+  if (peercaps) {
+    ret = gst_caps_intersect (ourcaps, peercaps);
+    gst_caps_unref (peercaps);
+  } else {
+    ret = gst_caps_copy (ourcaps);
+  }
+
+  GST_OBJECT_UNLOCK (self);
+
+  if (self->fields == GST_DEINTERLACE2_ALL) {
+    for (len = gst_caps_get_size (ret); len > 0; len--) {
+      GstStructure *s = gst_caps_get_structure (ret, len - 1);
+      const GValue *val;
+
+      val = gst_structure_get_value (s, "framerate");
+      if (!val)
+        continue;
+
+      if (G_VALUE_TYPE (val) == GST_TYPE_FRACTION) {
+        gint n, d;
+
+        n = gst_value_get_fraction_numerator (val);
+        d = gst_value_get_fraction_denominator (val);
+
+        if (!gst_fraction_double (&n, &d, pad != self->srcpad)) {
+          goto error;
+        }
+
+        gst_structure_set (s, "framerate", GST_TYPE_FRACTION, n, d, NULL);
+      } else if (G_VALUE_TYPE (val) == GST_TYPE_FRACTION_RANGE) {
+        const GValue *min, *max;
+        GValue nrange = { 0, }, nmin = {
+        0,}, nmax = {
+        0,};
+        gint n, d;
+
+        g_value_init (&nrange, GST_TYPE_FRACTION_RANGE);
+        g_value_init (&nmin, GST_TYPE_FRACTION);
+        g_value_init (&nmax, GST_TYPE_FRACTION);
+
+        min = gst_value_get_fraction_range_min (val);
+        max = gst_value_get_fraction_range_max (val);
+
+        n = gst_value_get_fraction_numerator (min);
+        d = gst_value_get_fraction_denominator (min);
+
+        if (!gst_fraction_double (&n, &d, pad != self->srcpad)) {
+          g_value_unset (&nrange);
+          g_value_unset (&nmax);
+          g_value_unset (&nmin);
+          goto error;
+        }
+
+        gst_value_set_fraction (&nmin, n, d);
+
+        n = gst_value_get_fraction_numerator (max);
+        d = gst_value_get_fraction_denominator (max);
+
+        if (!gst_fraction_double (&n, &d, pad != self->srcpad)) {
+          g_value_unset (&nrange);
+          g_value_unset (&nmax);
+          g_value_unset (&nmin);
+          goto error;
+        }
+
+        gst_value_set_fraction (&nmax, n, d);
+        gst_value_set_fraction_range (&nrange, &nmin, &nmax);
+
+        gst_structure_set_value (s, "framerate", &nrange);
+
+        g_value_unset (&nmin);
+        g_value_unset (&nmax);
+        g_value_unset (&nrange);
+      } else if (G_VALUE_TYPE (val) == GST_TYPE_LIST) {
+        const GValue *lval;
+        GValue nlist = { 0, };
+        GValue nval = { 0, };
+        gint i;
+
+        g_value_init (&nlist, GST_TYPE_LIST);
+        for (i = gst_value_list_get_size (val); i > 0; i--) {
+          gint n, d;
+
+          lval = gst_value_list_get_value (val, i);
+
+          if (G_VALUE_TYPE (lval) != GST_TYPE_FRACTION)
+            continue;
+
+          n = gst_value_get_fraction_numerator (lval);
+          d = gst_value_get_fraction_denominator (lval);
+
+          /* Double/Half the framerate but if this fails simply
+           * skip this value from the list */
+          if (!gst_fraction_double (&n, &d, pad != self->srcpad)) {
+            continue;
+          }
+
+          g_value_init (&nval, GST_TYPE_FRACTION);
+
+          gst_value_set_fraction (&nval, n, d);
+          gst_value_list_append_value (&nlist, &nval);
+          g_value_unset (&nval);
+        }
+        gst_structure_set_value (s, "framerate", &nlist);
+        g_value_unset (&nlist);
+      }
+    }
+  }
+
+  GST_DEBUG_OBJECT (pad, "Returning caps %" GST_PTR_FORMAT, ret);
+
+  return ret;
+
+error:
+  GST_ERROR_OBJECT (pad, "Unable to transform peer caps");
+  gst_caps_unref (ret);
+  return NULL;
+}
+
 static gboolean
 gst_deinterlace2_setcaps (GstPad * pad, GstCaps * caps)
 {
@@ -695,12 +1078,10 @@ gst_deinterlace2_setcaps (GstPad * pad, GstCaps * caps)
   if (self->fields == GST_DEINTERLACE2_ALL) {
     gint fps_n = self->frame_rate_n, fps_d = self->frame_rate_d;
 
-    othercaps = gst_caps_copy (caps);
+    if (!gst_fraction_double (&fps_n, &fps_d, otherpad != self->srcpad))
+      goto invalid_caps;
 
-    if (otherpad == self->srcpad)
-      fps_n *= 2;
-    else
-      fps_d *= 2;
+    othercaps = gst_caps_copy (caps);
 
     gst_caps_set_simple (othercaps, "framerate", GST_TYPE_FRACTION, fps_n,
         fps_d, NULL);
@@ -748,12 +1129,12 @@ done:
 
 invalid_caps:
   res = FALSE;
-  GST_ERROR_OBJECT (self, "Invalid caps: %" GST_PTR_FORMAT, caps);
+  GST_ERROR_OBJECT (pad, "Invalid caps: %" GST_PTR_FORMAT, caps);
   goto done;
 
 caps_not_accepted:
   res = FALSE;
-  GST_ERROR_OBJECT (self, "Caps not accepted: %" GST_PTR_FORMAT, othercaps);
+  GST_ERROR_OBJECT (pad, "Caps not accepted: %" GST_PTR_FORMAT, othercaps);
   gst_caps_unref (othercaps);
   goto done;
 }
