@@ -23,6 +23,7 @@
 
 #include "gstflvdemux.h"
 #include "gstflvparse.h"
+#include "gstflvmux.h"
 
 #include <string.h>
 
@@ -44,19 +45,14 @@ GST_STATIC_PAD_TEMPLATE ("video",
     GST_PAD_SOMETIMES,
     GST_STATIC_CAPS_ANY);
 
-static GstElementDetails flv_demux_details = {
-  "FLV Demuxer",
-  "Codec/Demuxer",
-  "Demux FLV feeds into digital streams",
-  "Julien Moutte <julien@moutte.net>"
-};
-
 GST_DEBUG_CATEGORY (flvdemux_debug);
 #define GST_CAT_DEFAULT flvdemux_debug
 
 GST_BOILERPLATE (GstFLVDemux, gst_flv_demux, GstElement, GST_TYPE_ELEMENT);
 
+/* 9 bytes of header + 4 bytes of first previous tag size */
 #define FLV_HEADER_SIZE 13
+/* 1 byte of tag type + 3 bytes of tag data size */
 #define FLV_TAG_TYPE_SIZE 4
 
 static void
@@ -176,7 +172,7 @@ gst_flv_demux_chain (GstPad * pad, GstBuffer * buffer)
     demux->offset = 0;
   }
 
-  if (G_UNLIKELY (demux->offset == 0)) {
+  if (G_UNLIKELY (demux->offset == 0 && GST_BUFFER_OFFSET (buffer) != 0)) {
     GST_DEBUG_OBJECT (demux, "offset was zero, synchronizing with buffer's");
     demux->offset = GST_BUFFER_OFFSET (buffer);
   }
@@ -184,11 +180,22 @@ gst_flv_demux_chain (GstPad * pad, GstBuffer * buffer)
   gst_adapter_push (demux->adapter, buffer);
 
 parse:
+  if (G_UNLIKELY (ret != GST_FLOW_OK)) {
+    if (ret == GST_FLOW_NOT_LINKED && (demux->audio_linked
+            || demux->video_linked)) {
+      ret = GST_FLOW_OK;
+    } else {
+      GST_DEBUG_OBJECT (demux, "got flow return %s", gst_flow_get_name (ret));
+      goto beach;
+    }
+  }
+
   if (G_UNLIKELY (demux->flushing)) {
     GST_DEBUG_OBJECT (demux, "we are now flushing, exiting parser loop");
     ret = GST_FLOW_WRONG_STATE;
     goto beach;
   }
+
   switch (demux->state) {
     case FLV_STATE_HEADER:
     {
@@ -294,6 +301,31 @@ beach:
   return ret;
 }
 
+static GstFlowReturn
+gst_flv_demux_pull_range (GstFLVDemux * demux, GstPad * pad, guint64 offset,
+    guint size, GstBuffer ** buffer)
+{
+  GstFlowReturn ret;
+
+  ret = gst_pad_pull_range (pad, offset, size, buffer);
+  if (G_UNLIKELY (ret != GST_FLOW_OK)) {
+    GST_WARNING_OBJECT (demux, "failed when pulling %d bytes", size);
+    *buffer = NULL;
+    return ret;
+  }
+
+  if (G_UNLIKELY (*buffer && GST_BUFFER_SIZE (*buffer) != size)) {
+    GST_WARNING_OBJECT (demux, "partial pull got %d when expecting %d",
+        GST_BUFFER_SIZE (*buffer), size);
+    gst_buffer_unref (*buffer);
+    ret = GST_FLOW_UNEXPECTED;
+    *buffer = NULL;
+    return ret;
+  }
+
+  return ret;
+}
+
 static GstFlowReturn
 gst_flv_demux_pull_tag (GstPad * pad, GstFLVDemux * demux)
 {
@@ -304,20 +336,9 @@ gst_flv_demux_pull_tag (GstPad * pad, GstFLVDemux * demux)
   demux->cur_tag_offset = demux->offset;
 
   /* Get the first 4 bytes to identify tag type and size */
-  ret = gst_pad_pull_range (pad, demux->offset, FLV_TAG_TYPE_SIZE, &buffer);
-  if (G_UNLIKELY (ret != GST_FLOW_OK)) {
-    GST_WARNING_OBJECT (demux, "failed when pulling %d bytes",
-        FLV_TAG_TYPE_SIZE);
+  if (G_UNLIKELY ((ret = gst_flv_demux_pull_range (demux, pad, demux->offset,
+                  FLV_TAG_TYPE_SIZE, &buffer)) != GST_FLOW_OK))
     goto beach;
-  }
-
-  if (G_UNLIKELY (buffer && GST_BUFFER_SIZE (buffer) != FLV_TAG_TYPE_SIZE)) {
-    GST_WARNING_OBJECT (demux, "partial pull got %d when expecting %d",
-        GST_BUFFER_SIZE (buffer), FLV_TAG_TYPE_SIZE);
-    gst_buffer_unref (buffer);
-    ret = GST_FLOW_UNEXPECTED;
-    goto beach;
-  }
 
   /* Identify tag type */
   ret = gst_flv_parse_tag_type (demux, GST_BUFFER_DATA (buffer),
@@ -329,21 +350,9 @@ gst_flv_demux_pull_tag (GstPad * pad, GstFLVDemux * demux)
   demux->offset += FLV_TAG_TYPE_SIZE;
 
   /* Pull the whole tag */
-  ret = gst_pad_pull_range (pad, demux->offset, demux->tag_size, &buffer);
-  if (G_UNLIKELY (ret != GST_FLOW_OK)) {
-    GST_WARNING_OBJECT (demux,
-        "failed when pulling %" G_GUINT64_FORMAT " bytes", demux->tag_size);
+  if (G_UNLIKELY ((ret = gst_flv_demux_pull_range (demux, pad, demux->offset,
+                  demux->tag_size, &buffer)) != GST_FLOW_OK))
     goto beach;
-  }
-
-  if (G_UNLIKELY (buffer && GST_BUFFER_SIZE (buffer) != demux->tag_size)) {
-    GST_WARNING_OBJECT (demux,
-        "partial pull got %d when expecting %" G_GUINT64_FORMAT,
-        GST_BUFFER_SIZE (buffer), demux->tag_size);
-    gst_buffer_unref (buffer);
-    ret = GST_FLOW_UNEXPECTED;
-    goto beach;
-  }
 
   switch (demux->state) {
     case FLV_STATE_TAG_VIDEO:
@@ -394,19 +403,9 @@ gst_flv_demux_pull_header (GstPad * pad, GstFLVDemux * demux)
   GstFlowReturn ret = GST_FLOW_OK;
 
   /* Get the first 9 bytes */
-  ret = gst_pad_pull_range (pad, demux->offset, FLV_HEADER_SIZE, &buffer);
-  if (G_UNLIKELY (ret != GST_FLOW_OK)) {
-    GST_WARNING_OBJECT (demux, "failed when pulling %d bytes", FLV_HEADER_SIZE);
+  if (G_UNLIKELY ((ret = gst_flv_demux_pull_range (demux, pad, demux->offset,
+                  FLV_HEADER_SIZE, &buffer)) != GST_FLOW_OK))
     goto beach;
-  }
-
-  if (G_UNLIKELY (buffer && GST_BUFFER_SIZE (buffer) != FLV_HEADER_SIZE)) {
-    GST_WARNING_OBJECT (demux, "partial pull got %d when expecting %d",
-        GST_BUFFER_SIZE (buffer), FLV_HEADER_SIZE);
-    gst_buffer_unref (buffer);
-    ret = GST_FLOW_UNEXPECTED;
-    goto beach;
-  }
 
   ret = gst_flv_parse_header (demux, GST_BUFFER_DATA (buffer),
       GST_BUFFER_SIZE (buffer));
@@ -444,6 +443,38 @@ gst_flv_demux_loop (GstPad * pad)
         break;
       default:
         ret = gst_flv_demux_pull_header (pad, demux);
+
+        /* If we parsed the header successfully try to get an
+         * approximate duration by looking at the last tag's timestamp */
+        if (ret == GST_FLOW_OK) {
+          gint64 size;
+          GstFormat fmt = GST_FORMAT_BYTES;
+
+          if (gst_pad_query_peer_duration (pad, &fmt, &size) &&
+              fmt == GST_FORMAT_BYTES && size != -1 && size > FLV_HEADER_SIZE) {
+            GstBuffer *buffer;
+
+            if (gst_flv_demux_pull_range (demux, pad, size - 4, 4,
+                    &buffer) == GST_FLOW_OK) {
+              guint32 prev_tag_size =
+                  GST_READ_UINT32_BE (GST_BUFFER_DATA (buffer));
+
+              gst_buffer_unref (buffer);
+
+              if (size - 4 - prev_tag_size > FLV_HEADER_SIZE &&
+                  prev_tag_size >= 8 &&
+                  gst_flv_demux_pull_range (demux, pad,
+                      size - prev_tag_size - 4, prev_tag_size,
+                      &buffer) == GST_FLOW_OK) {
+                demux->duration =
+                    gst_flv_parse_tag_timestamp (demux,
+                    GST_BUFFER_DATA (buffer), GST_BUFFER_SIZE (buffer));
+
+                gst_buffer_unref (buffer);
+              }
+            }
+          }
+        }
     }
 
     /* pause if something went wrong */
@@ -1157,7 +1188,10 @@ gst_flv_demux_base_init (gpointer g_class)
       gst_static_pad_template_get (&audio_src_template));
   gst_element_class_add_pad_template (element_class,
       gst_static_pad_template_get (&video_src_template));
-  gst_element_class_set_details (element_class, &flv_demux_details);
+  gst_element_class_set_details_simple (element_class, "FLV Demuxer",
+      "Codec/Demuxer",
+      "Demux FLV feeds into digital streams",
+      "Julien Moutte <julien@moutte.net>");
 }
 
 static void
@@ -1207,12 +1241,14 @@ plugin_init (GstPlugin * plugin)
   GST_DEBUG_CATEGORY_INIT (flvdemux_debug, "flvdemux", 0, "FLV demuxer");
 
   if (!gst_element_register (plugin, "flvdemux", GST_RANK_PRIMARY,
-          gst_flv_demux_get_type ()))
+          gst_flv_demux_get_type ()) ||
+      !gst_element_register (plugin, "flvmux", GST_RANK_NONE,
+          gst_flv_mux_get_type ()))
     return FALSE;
 
   return TRUE;
 }
 
 GST_PLUGIN_DEFINE (GST_VERSION_MAJOR, GST_VERSION_MINOR,
-    "flvdemux", "Element demuxing FLV stream",
+    "flv", "FLV muxing and demuxing plugin",
     plugin_init, VERSION, "LGPL", GST_PACKAGE_NAME, GST_PACKAGE_ORIGIN)
