@@ -406,15 +406,19 @@ gst_flv_parse_audio_negotiate (GstFLVDemux * demux, guint32 codec_tag,
       codec_name = "Shockwave ADPCM";
       break;
     case 2:
+    case 14:
       caps = gst_caps_new_simple ("audio/mpeg",
           "mpegversion", G_TYPE_INT, 1, "layer", G_TYPE_INT, 3, NULL);
       codec_name = "MPEG 1 Audio, Layer 3 (MP3)";
       break;
     case 0:
     case 3:
+      /* Assuming little endian for 0 (aka endianness of the
+       * system on which the file was created) as most people
+       * are probably using little endian machines */
       caps = gst_caps_new_simple ("audio/x-raw-int",
-          "endianness", G_TYPE_INT, G_BYTE_ORDER,
-          "signed", G_TYPE_BOOLEAN, TRUE,
+          "endianness", G_TYPE_INT, G_LITTLE_ENDIAN,
+          "signed", G_TYPE_BOOLEAN, (width == 8) ? FALSE : TRUE,
           "width", G_TYPE_INT, width, "depth", G_TYPE_INT, width, NULL);
       codec_name = "Raw Audio";
       break;
@@ -429,6 +433,14 @@ gst_flv_parse_audio_negotiate (GstFLVDemux * demux, guint32 codec_tag,
           "mpegversion", G_TYPE_INT, 4, NULL);
       codec_name = "AAC";
       break;
+    case 7:
+      caps = gst_caps_new_simple ("audio/x-alaw", NULL);
+      codec_name = "A-Law";
+      break;
+    case 8:
+      caps = gst_caps_new_simple ("audio/x-mulaw", NULL);
+      codec_name = "Mu-Law";
+      break;
     default:
       GST_WARNING_OBJECT (demux, "unsupported audio codec tag %u", codec_tag);
   }
@@ -523,6 +535,12 @@ gst_flv_parse_tag_audio (GstFLVDemux * demux, const guint8 * data,
     codec_data = 1;
   }
 
+  /* codec tags with special rates */
+  if (codec_tag == 5 || codec_tag == 14)
+    rate = 8000;
+  else if (codec_tag == 4)
+    rate = 16000;
+
   GST_LOG_OBJECT (demux, "audio tag with %d channels, %dHz sampling rate, "
       "%d bits width, codec tag %u (flags %02X)", channels, rate, width,
       codec_tag, flags);
@@ -530,16 +548,15 @@ gst_flv_parse_tag_audio (GstFLVDemux * demux, const guint8 * data,
   /* If we don't have our audio pad created, then create it. */
   if (G_UNLIKELY (!demux->audio_pad)) {
 
-    demux->audio_pad = gst_pad_new ("audio", GST_PAD_SRC);
+    demux->audio_pad =
+        gst_pad_new_from_template (gst_element_class_get_pad_template
+        (GST_ELEMENT_GET_CLASS (demux), "audio"), "audio");
     if (G_UNLIKELY (!demux->audio_pad)) {
       GST_WARNING_OBJECT (demux, "failed creating audio pad");
       ret = GST_FLOW_ERROR;
       goto beach;
     }
 
-    /* Make it active */
-    gst_pad_set_active (demux->audio_pad, TRUE);
-
     /* Negotiate caps */
     if (!gst_flv_parse_audio_negotiate (demux, codec_tag, rate, channels,
             width)) {
@@ -560,6 +577,11 @@ gst_flv_parse_tag_audio (GstFLVDemux * demux, const guint8 * data,
     gst_pad_set_event_function (demux->audio_pad,
         GST_DEBUG_FUNCPTR (gst_flv_demux_src_event));
 
+    gst_pad_use_fixed_caps (demux->audio_pad);
+
+    /* Make it active */
+    gst_pad_set_active (demux->audio_pad, TRUE);
+
     /* We need to set caps before adding */
     gst_element_add_pad (GST_ELEMENT (demux),
         gst_object_ref (demux->audio_pad));
@@ -603,9 +625,10 @@ gst_flv_parse_tag_audio (GstFLVDemux * demux, const guint8 * data,
   }
 
   /* Create buffer from pad */
-  ret = gst_pad_alloc_buffer (demux->audio_pad, GST_BUFFER_OFFSET_NONE,
-      demux->tag_data_size - codec_data, GST_PAD_CAPS (demux->audio_pad),
-      &buffer);
+  ret =
+      gst_pad_alloc_buffer_and_set_caps (demux->audio_pad,
+      GST_BUFFER_OFFSET_NONE, demux->tag_data_size - codec_data,
+      GST_PAD_CAPS (demux->audio_pad), &buffer);
   if (G_UNLIKELY (ret != GST_FLOW_OK)) {
     GST_WARNING_OBJECT (demux, "failed allocating a %" G_GUINT64_FORMAT
         " bytes buffer: %s", demux->tag_data_size, gst_flow_get_name (ret));
@@ -653,6 +676,19 @@ gst_flv_parse_tag_audio (GstFLVDemux * demux, const guint8 * data,
   GST_BUFFER_OFFSET (buffer) = demux->audio_offset++;
   GST_BUFFER_OFFSET_END (buffer) = demux->audio_offset;
 
+  /* Only add audio frames to the index if we have no video */
+  if (!demux->has_video) {
+    if (demux->index) {
+      GST_LOG_OBJECT (demux, "adding association %" GST_TIME_FORMAT "-> %"
+          G_GUINT64_FORMAT, GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buffer)),
+          demux->cur_tag_offset);
+      gst_index_add_association (demux->index, demux->index_id,
+          GST_ASSOCIATION_FLAG_KEY_UNIT,
+          GST_FORMAT_TIME, GST_BUFFER_TIMESTAMP (buffer),
+          GST_FORMAT_BYTES, demux->cur_tag_offset, NULL);
+    }
+  }
+
   if (G_UNLIKELY (demux->audio_need_discont)) {
     GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_DISCONT);
     demux->audio_need_discont = FALSE;
@@ -710,10 +746,13 @@ gst_flv_parse_video_negotiate (GstFLVDemux * demux, guint32 codec_tag)
       caps = gst_caps_new_simple ("video/x-flash-screen", NULL);
       codec_name = "Flash Screen Video";
     case 4:
-    case 5:
       caps = gst_caps_new_simple ("video/x-vp6-flash", NULL);
       codec_name = "On2 VP6 Video";
       break;
+    case 5:
+      caps = gst_caps_new_simple ("video/x-vp6-alpha", NULL);
+      codec_name = "On2 VP6 Video with alpha channel";
+      break;
     case 7:
       caps = gst_caps_new_simple ("video/x-h264", NULL);
       codec_name = "H.264/AVC Video";
@@ -802,14 +841,14 @@ gst_flv_parse_tag_video (GstFLVDemux * demux, const guint8 * data,
 
   /* If we don't have our video pad created, then create it. */
   if (G_UNLIKELY (!demux->video_pad)) {
-    demux->video_pad = gst_pad_new ("video", GST_PAD_SRC);
+    demux->video_pad =
+        gst_pad_new_from_template (gst_element_class_get_pad_template
+        (GST_ELEMENT_GET_CLASS (demux), "video"), "video");
     if (G_UNLIKELY (!demux->video_pad)) {
       GST_WARNING_OBJECT (demux, "failed creating video pad");
       ret = GST_FLOW_ERROR;
       goto beach;
     }
-    /* Make it active */
-    gst_pad_set_active (demux->video_pad, TRUE);
 
     if (!gst_flv_parse_video_negotiate (demux, codec_tag)) {
       gst_object_unref (demux->video_pad);
@@ -833,6 +872,11 @@ gst_flv_parse_tag_video (GstFLVDemux * demux, const guint8 * data,
     gst_pad_set_event_function (demux->video_pad,
         GST_DEBUG_FUNCPTR (gst_flv_demux_src_event));
 
+    gst_pad_use_fixed_caps (demux->video_pad);
+
+    /* Make it active */
+    gst_pad_set_active (demux->video_pad, TRUE);
+
     /* We need to set caps before adding */
     gst_element_add_pad (GST_ELEMENT (demux),
         gst_object_ref (demux->video_pad));
@@ -879,9 +923,10 @@ gst_flv_parse_tag_video (GstFLVDemux * demux, const guint8 * data,
   }
 
   /* Create buffer from pad */
-  ret = gst_pad_alloc_buffer (demux->video_pad, GST_BUFFER_OFFSET_NONE,
-      demux->tag_data_size - codec_data, GST_PAD_CAPS (demux->video_pad),
-      &buffer);
+  ret =
+      gst_pad_alloc_buffer_and_set_caps (demux->video_pad,
+      GST_BUFFER_OFFSET_NONE, demux->tag_data_size - codec_data,
+      GST_PAD_CAPS (demux->video_pad), &buffer);
   if (G_UNLIKELY (ret != GST_FLOW_OK)) {
     GST_WARNING_OBJECT (demux, "failed allocating a %" G_GUINT64_FORMAT
         " bytes buffer: %s", demux->tag_data_size, gst_flow_get_name (ret));
@@ -931,6 +976,15 @@ gst_flv_parse_tag_video (GstFLVDemux * demux, const guint8 * data,
 
   if (!keyframe) {
     GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_DELTA_UNIT);
+    if (demux->index) {
+      GST_LOG_OBJECT (demux, "adding association %" GST_TIME_FORMAT "-> %"
+          G_GUINT64_FORMAT, GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buffer)),
+          demux->cur_tag_offset);
+      gst_index_add_association (demux->index, demux->index_id,
+          GST_ASSOCIATION_FLAG_NONE,
+          GST_FORMAT_TIME, GST_BUFFER_TIMESTAMP (buffer),
+          GST_FORMAT_BYTES, demux->cur_tag_offset, NULL);
+    }
   } else {
     if (demux->index) {
       GST_LOG_OBJECT (demux, "adding association %" GST_TIME_FORMAT "-> %"
@@ -985,6 +1039,37 @@ beach:
   return ret;
 }
 
+GstClockTime
+gst_flv_parse_tag_timestamp (GstFLVDemux * demux, const guint8 * data,
+    size_t data_size)
+{
+  guint32 pts = 0, pts_ext = 0;
+
+  if (data[0] != 9 && data[0] != 8 && data[0] != 18) {
+    GST_WARNING_OBJECT (demux, "Unsupported tag type %u", data[0]);
+    return GST_CLOCK_TIME_NONE;
+  }
+
+  if (FLV_GET_BEUI24 (data + 1, data_size - 1) != data_size - 11) {
+    GST_WARNING_OBJECT (demux, "Invalid tag");
+    return GST_CLOCK_TIME_NONE;
+  }
+
+  data += 4;
+
+  GST_LOG_OBJECT (demux, "pts bytes %02X %02X %02X %02X", data[0], data[1],
+      data[2], data[3]);
+
+  /* Grab timestamp of tag tag */
+  pts = FLV_GET_BEUI24 (data, data_size);
+  /* read the pts extension to 32 bits integer */
+  pts_ext = GST_READ_UINT8 (data + 3);
+  /* Combine them */
+  pts |= pts_ext << 24;
+
+  return pts * GST_MSECOND;
+}
+
 GstFlowReturn
 gst_flv_parse_tag_type (GstFLVDemux * demux, const guint8 * data,
     size_t data_size)
