@@ -119,6 +119,9 @@
 
 #include "gstrtpbin-marshal.h"
 #include "gstrtpbin.h"
+#include "rtpsession.h"
+#include "gstrtpsession.h"
+#include "gstrtpjitterbuffer.h"
 
 GST_DEBUG_CATEGORY_STATIC (gst_rtp_bin_debug);
 #define GST_CAT_DEFAULT gst_rtp_bin_debug
@@ -226,6 +229,7 @@ enum
 {
   SIGNAL_REQUEST_PT_MAP,
   SIGNAL_CLEAR_PT_MAP,
+  SIGNAL_GET_INTERNAL_SESSION,
 
   SIGNAL_ON_NEW_SSRC,
   SIGNAL_ON_SSRC_COLLISION,
@@ -235,6 +239,7 @@ enum
   SIGNAL_ON_BYE_SSRC,
   SIGNAL_ON_BYE_TIMEOUT,
   SIGNAL_ON_TIMEOUT,
+  SIGNAL_ON_SENDER_TIMEOUT,
   LAST_SIGNAL
 };
 
@@ -317,11 +322,11 @@ struct _GstRtpBinStream
   gint64 unix_delta;
 
   /* for lip-sync */
+  guint64 last_clock_base;
   guint64 clock_base;
   guint64 clock_base_time;
   gint clock_rate;
   gint64 ts_offset;
-  gint64 prev_ts_offset;
   gint last_pt;
 };
 
@@ -397,6 +402,23 @@ find_session_by_id (GstRtpBin * rtpbin, gint id)
   return NULL;
 }
 
+/* find a session with the given request pad. Must be called with RTP_BIN_LOCK */
+static GstRtpBinSession *
+find_session_by_pad (GstRtpBin * rtpbin, GstPad * pad)
+{
+  GSList *walk;
+
+  for (walk = rtpbin->sessions; walk; walk = g_slist_next (walk)) {
+    GstRtpBinSession *sess = (GstRtpBinSession *) walk->data;
+
+    if ((sess->recv_rtp_sink == pad) ||
+        (sess->recv_rtcp_sink == pad) ||
+        (sess->send_rtp_sink == pad) || (sess->send_rtcp_src == pad))
+      return sess;
+  }
+  return NULL;
+}
+
 static void
 on_new_ssrc (GstElement * session, guint32 ssrc, GstRtpBinSession * sess)
 {
@@ -453,6 +475,13 @@ on_timeout (GstElement * session, guint32 ssrc, GstRtpBinSession * sess)
       sess->id, ssrc);
 }
 
+static void
+on_sender_timeout (GstElement * session, guint32 ssrc, GstRtpBinSession * sess)
+{
+  g_signal_emit (sess->bin, gst_rtp_bin_signals[SIGNAL_ON_SENDER_TIMEOUT], 0,
+      sess->id, ssrc);
+}
+
 /* create a session with the given id.  Must be called with RTP_BIN_LOCK */
 static GstRtpBinSession *
 create_session (GstRtpBin * rtpbin, gint id)
@@ -505,6 +534,8 @@ create_session (GstRtpBin * rtpbin, gint id)
   g_signal_connect (sess->session, "on-bye-timeout",
       (GCallback) on_bye_timeout, sess);
   g_signal_connect (sess->session, "on-timeout", (GCallback) on_timeout, sess);
+  g_signal_connect (sess->session, "on-sender-timeout",
+      (GCallback) on_sender_timeout, sess);
 
   /* FIXME, change state only to what's needed */
   gst_bin_add (GST_BIN_CAST (rtpbin), session);
@@ -535,23 +566,33 @@ free_session (GstRtpBinSession * sess)
 
   bin = sess->bin;
 
+  GST_DEBUG_OBJECT (bin, "freeing session %p", sess);
+
   gst_element_set_state (sess->session, GST_STATE_NULL);
   gst_element_set_state (sess->demux, GST_STATE_NULL);
 
-  if (sess->recv_rtp_sink != NULL)
+  if (sess->recv_rtp_sink != NULL) {
     gst_element_release_request_pad (sess->session, sess->recv_rtp_sink);
+    gst_object_unref (sess->recv_rtp_sink);
+  }
   if (sess->recv_rtp_src != NULL)
     gst_object_unref (sess->recv_rtp_src);
-  if (sess->recv_rtcp_sink != NULL)
+  if (sess->recv_rtcp_sink != NULL) {
     gst_element_release_request_pad (sess->session, sess->recv_rtcp_sink);
+    gst_object_unref (sess->recv_rtcp_sink);
+  }
   if (sess->sync_src != NULL)
     gst_object_unref (sess->sync_src);
-  if (sess->send_rtp_sink != NULL)
+  if (sess->send_rtp_sink != NULL) {
     gst_element_release_request_pad (sess->session, sess->send_rtp_sink);
+    gst_object_unref (sess->send_rtp_sink);
+  }
   if (sess->send_rtp_src != NULL)
     gst_object_unref (sess->send_rtp_src);
-  if (sess->send_rtcp_src != NULL)
+  if (sess->send_rtcp_src != NULL) {
     gst_element_release_request_pad (sess->session, sess->send_rtcp_src);
+    gst_object_unref (sess->send_rtcp_src);
+  }
 
   gst_bin_remove (GST_BIN_CAST (bin), sess->session);
   gst_bin_remove (GST_BIN_CAST (bin), sess->demux);
@@ -618,11 +659,24 @@ get_pt_map (GstRtpBinSession * session, guint pt)
   g_value_init (&ret, GST_TYPE_CAPS);
   g_value_set_boxed (&ret, NULL);
 
+  GST_RTP_SESSION_UNLOCK (session);
+
   g_signal_emitv (args, gst_rtp_bin_signals[SIGNAL_REQUEST_PT_MAP], 0, &ret);
 
+  GST_RTP_SESSION_LOCK (session);
+
   g_value_unset (&args[0]);
   g_value_unset (&args[1]);
   g_value_unset (&args[2]);
+
+  /* look in the cache again because we let the lock go */
+  caps = g_hash_table_lookup (session->ptmap, GINT_TO_POINTER (pt));
+  if (caps) {
+    gst_caps_ref (caps);
+    g_value_unset (&ret);
+    goto done;
+  }
+
   caps = (GstCaps *) g_value_dup_boxed (&ret);
   g_value_unset (&ret);
   if (!caps)
@@ -682,6 +736,25 @@ gst_rtp_bin_clear_pt_map (GstRtpBin * bin)
   GST_RTP_BIN_UNLOCK (bin);
 }
 
+static RTPSession *
+gst_rtp_bin_get_internal_session (GstRtpBin * bin, guint session_id)
+{
+  RTPSession *internal_session = NULL;
+  GstRtpBinSession *session;
+
+  GST_RTP_BIN_LOCK (bin);
+  GST_DEBUG_OBJECT (bin, "retrieving internal RTPSession object, index: %d",
+      session_id);
+  session = find_session_by_id (bin, (gint) session_id);
+  if (session) {
+    g_object_get (session->session, "internal-session", &internal_session,
+        NULL);
+  }
+  GST_RTP_BIN_UNLOCK (bin);
+
+  return internal_session;
+}
+
 static void
 gst_rtp_bin_propagate_property_to_jitterbuffer (GstRtpBin * bin,
     const gchar * name, const GValue * value)
@@ -807,7 +880,9 @@ gst_rtp_bin_associate (GstRtpBin * bin, GstRtpBinStream * stream, guint8 len,
       goto no_clock_rate;
   }
 
-  /* map last RTP time to local timeline using our clock-base */
+  /* take the extended rtptime we found in the SR packet and map it to the
+   * local rtptime. The local rtp time is used to construct timestamps on the
+   * buffers. */
   stream->local_rtp = stream->last_extrtptime - stream->clock_base;
 
   GST_DEBUG_OBJECT (bin,
@@ -815,12 +890,16 @@ gst_rtp_bin_associate (GstRtpBin * bin, GstRtpBinStream * stream, guint8 len,
       ", local RTP %" G_GUINT64_FORMAT ", clock-rate %d", stream->clock_base,
       stream->last_extrtptime, stream->local_rtp, stream->clock_rate);
 
-  /* calculate local NTP time in gstreamer timestamp */
+  /* calculate local NTP time in gstreamer timestamp, we essentially perform the
+   * same conversion that a jitterbuffer would use to convert an rtp timestamp
+   * into a corresponding gstreamer timestamp. */
   stream->local_unix =
       gst_util_uint64_scale_int (stream->local_rtp, GST_SECOND,
       stream->clock_rate);
   stream->local_unix += stream->clock_base_time;
-  /* calculate delta between server and receiver */
+  /* calculate delta between server and receiver. last_unix is created by
+   * converting the ntptime in the last SR packet to a gstreamer timestamp. This
+   * delta expresses the difference to our timeline and the server timeline. */
   stream->unix_delta = stream->last_unix - stream->local_unix;
 
   GST_DEBUG_OBJECT (bin,
@@ -828,17 +907,28 @@ gst_rtp_bin_associate (GstRtpBin * bin, GstRtpBinStream * stream, guint8 len,
       ", delta %" G_GINT64_FORMAT, stream->local_unix, stream->last_unix,
       stream->unix_delta);
 
-  /* recalc inter stream playout offset, but only if there are more than one
+  /* recalc inter stream playout offset, but only if there is more than one
    * stream. */
   if (client->nstreams > 1) {
     gint64 min;
 
-    /* calculate the min of all deltas */
+    /* calculate the min of all deltas, ignoring streams that did not yet have a
+     * valid unix_delta because we did not yet receive an SR packet for those
+     * streams. 
+     * We calculate the mininum because we would like to only apply positive
+     * offsets to streams, delaying their playback instead of trying to speed up
+     * other streams (which might be imposible when we have to create negative
+     * latencies).
+     * The stream that has the smalest diff is selected as the reference stream,
+     * all other streams will have a positive offset to this difference. */
     min = G_MAXINT64;
     for (walk = client->streams; walk; walk = g_slist_next (walk)) {
       GstRtpBinStream *ostream = (GstRtpBinStream *) walk->data;
 
-      if (ostream->unix_delta && ostream->unix_delta < min)
+      if (!ostream->have_sync)
+        continue;
+
+      if (ostream->unix_delta < min)
         min = ostream->unix_delta;
     }
 
@@ -848,34 +938,47 @@ gst_rtp_bin_associate (GstRtpBin * bin, GstRtpBinStream * stream, guint8 len,
     /* calculate offsets for each stream */
     for (walk = client->streams; walk; walk = g_slist_next (walk)) {
       GstRtpBinStream *ostream = (GstRtpBinStream *) walk->data;
+      gint64 prev_ts_offset;
 
-      if (ostream->unix_delta == 0)
+      /* ignore streams for which we didn't receive an SR packet yet, we
+       * can't synchronize them yet. We can however sync other streams just
+       * fine. */
+      if (!ostream->have_sync)
         continue;
 
+      /* calculate offset to our reference stream, this should always give a
+       * positive number. */
       ostream->ts_offset = ostream->unix_delta - min;
 
+      g_object_get (ostream->buffer, "ts-offset", &prev_ts_offset, NULL);
+
       /* delta changed, see how much */
-      if (ostream->prev_ts_offset != ostream->ts_offset) {
+      if (prev_ts_offset != ostream->ts_offset) {
         gint64 diff;
 
-        if (ostream->prev_ts_offset > ostream->ts_offset)
-          diff = ostream->prev_ts_offset - ostream->ts_offset;
+        if (prev_ts_offset > ostream->ts_offset)
+          diff = prev_ts_offset - ostream->ts_offset;
         else
-          diff = ostream->ts_offset - ostream->prev_ts_offset;
+          diff = ostream->ts_offset - prev_ts_offset;
 
-        /* only change diff when it changed more than 1 millisecond. This
+        GST_DEBUG_OBJECT (bin,
+            "ts-offset %" G_GUINT64_FORMAT ", prev %" G_GUINT64_FORMAT
+            ", diff: %" G_GINT64_FORMAT, ostream->ts_offset, prev_ts_offset,
+            diff);
+
+        /* only change diff when it changed more than 4 milliseconds. This
          * compensates for rounding errors in NTP to RTP timestamp
          * conversions */
-        if (diff > GST_MSECOND)
+        if (diff > 4 * GST_MSECOND && diff < (3 * GST_SECOND)) {
           g_object_set (ostream->buffer, "ts-offset", ostream->ts_offset, NULL);
-
-        ostream->prev_ts_offset = ostream->ts_offset;
+        }
       }
       GST_DEBUG_OBJECT (bin, "stream SSRC %08x, delta %" G_GINT64_FORMAT,
           ostream->ssrc, ostream->ts_offset);
     }
   }
   GST_RTP_BIN_UNLOCK (bin);
+
   return;
 
 no_clock_base:
@@ -916,6 +1019,8 @@ gst_rtp_bin_sync_chain (GstPad * pad, GstBuffer * buffer)
   guint32 rtptime;
   gboolean have_sr, have_sdes;
   gboolean more;
+  guint64 clock_base;
+  guint64 clock_base_time;
 
   stream = gst_pad_get_element_private (pad);
   bin = stream->bin;
@@ -925,6 +1030,21 @@ gst_rtp_bin_sync_chain (GstPad * pad, GstBuffer * buffer)
   if (!gst_rtcp_buffer_validate (buffer))
     goto invalid_rtcp;
 
+  /* get the last relation between the rtp timestamps and the gstreamer
+   * timestamps. We get this info directly from the jitterbuffer which
+   * constructs gstreamer timestamps from rtp timestamps and so it know exactly
+   * what the current situation is. */
+  gst_rtp_jitter_buffer_get_sync (GST_RTP_JITTER_BUFFER (stream->buffer),
+      &clock_base, &clock_base_time);
+
+  /* clock base changes when there is a huge gap in the timestamps or seqnum.
+   * When this happens we don't want to calculate the extended timestamp based
+   * on the previous one but reset the calculation. */
+  if (stream->last_clock_base != clock_base) {
+    stream->last_extrtptime = -1;
+    stream->last_clock_base = clock_base;
+  }
+
   have_sr = FALSE;
   have_sdes = FALSE;
   GST_RTCP_BUFFER_FOR_PACKETS (more, buffer, &packet) {
@@ -976,8 +1096,8 @@ gst_rtp_bin_sync_chain (GstPad * pad, GstBuffer * buffer)
             gst_rtcp_packet_sdes_get_entry (&packet, &type, &len, &data);
 
             if (type == GST_RTCP_SDES_CNAME) {
-              stream->clock_base = GST_BUFFER_OFFSET (buffer);
-              stream->clock_base_time = GST_BUFFER_OFFSET_END (buffer);
+              stream->clock_base = clock_base;
+              stream->clock_base_time = clock_base_time;
               /* associate the stream to CNAME */
               gst_rtp_bin_associate (bin, stream, len, data);
             }
@@ -1191,6 +1311,18 @@ gst_rtp_bin_class_init (GstRtpBinClass * klass)
       G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION, G_STRUCT_OFFSET (GstRtpBinClass,
           clear_pt_map), NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
       0, G_TYPE_NONE);
+  /**
+   * GstRtpBin::get-internal-session:
+   * @rtpbin: the object which received the signal
+   * @id: the session id
+   *
+   * Request the internal RTPSession object as #GObject in session @id.
+   */
+  gst_rtp_bin_signals[SIGNAL_GET_INTERNAL_SESSION] =
+      g_signal_new ("get-internal-session", G_TYPE_FROM_CLASS (klass),
+      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION, G_STRUCT_OFFSET (GstRtpBinClass,
+          get_internal_session), NULL, NULL, gst_rtp_bin_marshal_OBJECT__UINT,
+      RTP_TYPE_SESSION, 1, G_TYPE_UINT);
 
   /**
    * GstRtpBin::on-new-ssrc:
@@ -1297,6 +1429,19 @@ gst_rtp_bin_class_init (GstRtpBinClass * klass)
       G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (GstRtpBinClass, on_timeout),
       NULL, NULL, gst_rtp_bin_marshal_VOID__UINT_UINT, G_TYPE_NONE, 2,
       G_TYPE_UINT, G_TYPE_UINT);
+  /**
+   * GstRtpBin::on-sender-timeout:
+   * @rtpbin: the object which received the signal
+   * @session: the session
+   * @ssrc: the SSRC 
+   *
+   * Notify of a sender SSRC that has timed out and became a receiver
+   */
+  gst_rtp_bin_signals[SIGNAL_ON_SENDER_TIMEOUT] =
+      g_signal_new ("on-sender-timeout", G_TYPE_FROM_CLASS (klass),
+      G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (GstRtpBinClass, on_sender_timeout),
+      NULL, NULL, gst_rtp_bin_marshal_VOID__UINT_UINT, G_TYPE_NONE, 2,
+      G_TYPE_UINT, G_TYPE_UINT);
 
   g_object_class_install_property (gobject_class, PROP_SDES_CNAME,
       g_param_spec_string ("sdes-cname", "SDES CNAME",
@@ -1348,6 +1493,8 @@ gst_rtp_bin_class_init (GstRtpBinClass * klass)
   gstbin_class->handle_message = GST_DEBUG_FUNCPTR (gst_rtp_bin_handle_message);
 
   klass->clear_pt_map = GST_DEBUG_FUNCPTR (gst_rtp_bin_clear_pt_map);
+  klass->get_internal_session =
+      GST_DEBUG_FUNCPTR (gst_rtp_bin_get_internal_session);
 
   GST_DEBUG_CATEGORY_INIT (gst_rtp_bin_debug, "rtpbin", 0, "RTP bin");
 }
@@ -1381,9 +1528,11 @@ gst_rtp_bin_dispose (GObject * object)
 
   rtpbin = GST_RTP_BIN (object);
 
+  GST_DEBUG_OBJECT (object, "freeing sessions");
   g_slist_foreach (rtpbin->sessions, (GFunc) free_session, NULL);
   g_slist_free (rtpbin->sessions);
   rtpbin->sessions = NULL;
+  GST_DEBUG_OBJECT (object, "freeing clients");
   g_slist_foreach (rtpbin->clients, (GFunc) free_client, NULL);
   g_slist_free (rtpbin->clients);
   rtpbin->clients = NULL;
@@ -1834,7 +1983,8 @@ new_ssrc_pad_found (GstElement * element, guint ssrc, GstPad * pad,
 
   rtpbin = session->bin;
 
-  GST_DEBUG_OBJECT (rtpbin, "new SSRC pad %08x", ssrc);
+  GST_DEBUG_OBJECT (rtpbin, "new SSRC pad %08x, %s:%s", ssrc,
+      GST_DEBUG_PAD_NAME (pad));
 
   GST_RTP_BIN_SHUTDOWN_LOCK (rtpbin, shutdown);
 
@@ -1862,6 +2012,7 @@ new_ssrc_pad_found (GstElement * element, guint ssrc, GstPad * pad,
           gst_caps_to_string (caps), GST_DEBUG_PAD_NAME (pad));
     }
 
+    stream->last_clock_base = -1;
     if (gst_structure_get_uint (s, "clock-base", &val))
       stream->clock_base = val;
     else
@@ -2020,6 +2171,13 @@ link_failed:
   }
 }
 
+static void
+remove_recv_rtp (GstRtpBin * rtpbin, GstRtpBinSession * session, GstPad * pad)
+{
+  g_warning ("gstrtpbin: releasing pad %s:%s is not implemented",
+      GST_DEBUG_PAD_NAME (pad));
+}
+
 /* Create a pad for receiving RTCP for the session in @name. Must be called with
  * RTP_BIN_LOCK.
  */
@@ -2109,6 +2267,13 @@ link_failed:
   }
 }
 
+static void
+remove_recv_rtcp (GstRtpBin * rtpbin, GstRtpBinSession * session, GstPad * pad)
+{
+  g_warning ("gstrtpbin: releasing pad %s:%s is not implemented",
+      GST_DEBUG_PAD_NAME (pad));
+}
+
 /* Create a pad for sending RTP for the session in @name. Must be called with
  * RTP_BIN_LOCK.
  */
@@ -2197,6 +2362,13 @@ no_srcpad:
   }
 }
 
+static void
+remove_send_rtp (GstRtpBin * rtpbin, GstRtpBinSession * session, GstPad * pad)
+{
+  g_warning ("gstrtpbin: releasing pad %s:%s is not implemented",
+      GST_DEBUG_PAD_NAME (pad));
+}
+
 /* Create a pad for sending RTCP for the session in @name. Must be called with
  * RTP_BIN_LOCK.
  */
@@ -2257,6 +2429,13 @@ pad_failed:
   }
 }
 
+static void
+remove_rtcp (GstRtpBin * rtpbin, GstRtpBinSession * session, GstPad * pad)
+{
+  g_warning ("gstrtpbin: releasing pad %s:%s is not implemented",
+      GST_DEBUG_PAD_NAME (pad));
+}
+
 /* If the requested name is NULL we should create a name with
  * the session number assuming we want the lowest posible session
  * with a free pad like the template */
@@ -2299,6 +2478,7 @@ gst_rtp_bin_request_new_pad (GstElement * element,
   GstRtpBin *rtpbin;
   GstElementClass *klass;
   GstPad *result;
+
   gchar *pad_name = NULL;
 
   g_return_val_if_fail (templ != NULL, NULL);
@@ -2352,4 +2532,38 @@ wrong_template:
 static void
 gst_rtp_bin_release_pad (GstElement * element, GstPad * pad)
 {
+  GstRtpBinSession *session;
+  GstRtpBin *rtpbin;
+
+  g_return_if_fail (GST_IS_PAD (pad));
+  g_return_if_fail (GST_IS_RTP_BIN (element));
+
+  rtpbin = GST_RTP_BIN (element);
+
+  GST_RTP_BIN_LOCK (rtpbin);
+  if (!(session = find_session_by_pad (rtpbin, pad)))
+    goto unknown_pad;
+
+  if (session->recv_rtp_sink == pad) {
+    remove_recv_rtp (rtpbin, session, pad);
+  } else if (session->recv_rtcp_sink == pad) {
+    remove_recv_rtcp (rtpbin, session, pad);
+  } else if (session->send_rtp_sink == pad) {
+    remove_send_rtp (rtpbin, session, pad);
+  } else if (session->send_rtcp_src == pad) {
+    remove_rtcp (rtpbin, session, pad);
+  }
+
+  GST_RTP_BIN_UNLOCK (rtpbin);
+
+  return;
+
+  /* ERROR */
+unknown_pad:
+  {
+    GST_RTP_BIN_UNLOCK (rtpbin);
+    g_warning ("gstrtpbin: %s:%s is not one of our request pads",
+        GST_DEBUG_PAD_NAME (pad));
+    return;
+  }
 }
