@@ -23,7 +23,7 @@
 #include <gst/rtp/gstrtcpbuffer.h>
 #include <gst/netbuffer/gstnetbuffer.h>
 
-
+#include "gstrtpbin-marshal.h"
 #include "rtpsession.h"
 
 GST_DEBUG_CATEGORY_STATIC (rtp_session_debug);
@@ -32,6 +32,7 @@ GST_DEBUG_CATEGORY_STATIC (rtp_session_debug);
 /* signals and args */
 enum
 {
+  SIGNAL_GET_SOURCE_BY_SSRC,
   SIGNAL_ON_NEW_SSRC,
   SIGNAL_ON_SSRC_COLLISION,
   SIGNAL_ON_SSRC_VALIDATED,
@@ -40,6 +41,7 @@ enum
   SIGNAL_ON_BYE_SSRC,
   SIGNAL_ON_BYE_TIMEOUT,
   SIGNAL_ON_TIMEOUT,
+  SIGNAL_ON_SENDER_TIMEOUT,
   LAST_SIGNAL
 };
 
@@ -116,6 +118,19 @@ rtp_session_class_init (RTPSessionClass * klass)
   gobject_class->set_property = rtp_session_set_property;
   gobject_class->get_property = rtp_session_get_property;
 
+  /**
+   * RTPSession::get-source-by-ssrc:
+   * @session: the object which received the signal
+   * @ssrc: the SSRC of the RTPSource
+   *
+   * Request the #RTPSource object with SSRC @ssrc in @session.
+   */
+  rtp_session_signals[SIGNAL_GET_SOURCE_BY_SSRC] =
+      g_signal_new ("get-source-by-ssrc", G_TYPE_FROM_CLASS (klass),
+      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION, G_STRUCT_OFFSET (RTPSessionClass,
+          get_source_by_ssrc), NULL, NULL, gst_rtp_bin_marshal_OBJECT__UINT,
+      RTP_TYPE_SOURCE, 1, G_TYPE_UINT);
+
   /**
    * RTPSession::on-new-ssrc:
    * @session: the object which received the signal
@@ -212,6 +227,18 @@ rtp_session_class_init (RTPSessionClass * klass)
       G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (RTPSessionClass, on_timeout),
       NULL, NULL, g_cclosure_marshal_VOID__OBJECT, G_TYPE_NONE, 1,
       RTP_TYPE_SOURCE);
+  /**
+   * RTPSession::on-sender-timeout:
+   * @session: the object which received the signal
+   * @src: the RTPSource that timed out
+   *
+   * Notify of an SSRC that was a sender but timed out and became a receiver.
+   */
+  rtp_session_signals[SIGNAL_ON_SENDER_TIMEOUT] =
+      g_signal_new ("on-sender-timeout", G_TYPE_FROM_CLASS (klass),
+      G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (RTPSessionClass, on_sender_timeout),
+      NULL, NULL, g_cclosure_marshal_VOID__OBJECT, G_TYPE_NONE, 1,
+      RTP_TYPE_SOURCE);
 
   g_object_class_install_property (gobject_class, PROP_INTERNAL_SOURCE,
       g_param_spec_object ("internal-source", "Internal Source",
@@ -273,6 +300,9 @@ rtp_session_class_init (RTPSessionClass * klass)
           "The number of active sources in the session", 0, G_MAXUINT,
           DEFAULT_NUM_ACTIVE_SOURCES, G_PARAM_READABLE));
 
+  klass->get_source_by_ssrc =
+      GST_DEBUG_FUNCPTR (rtp_session_get_source_by_ssrc);
+
   GST_DEBUG_CATEGORY_INIT (rtp_session_debug, "rtpsession", 0, "RTP Session");
 }
 
@@ -449,68 +479,95 @@ rtp_session_get_property (GObject * object, guint prop_id,
 static void
 on_new_ssrc (RTPSession * sess, RTPSource * source)
 {
+  g_object_ref (source);
   RTP_SESSION_UNLOCK (sess);
   g_signal_emit (sess, rtp_session_signals[SIGNAL_ON_NEW_SSRC], 0, source);
   RTP_SESSION_LOCK (sess);
+  g_object_unref (source);
 }
 
 static void
 on_ssrc_collision (RTPSession * sess, RTPSource * source)
 {
+  g_object_ref (source);
   RTP_SESSION_UNLOCK (sess);
   g_signal_emit (sess, rtp_session_signals[SIGNAL_ON_SSRC_COLLISION], 0,
       source);
   RTP_SESSION_LOCK (sess);
+  g_object_unref (source);
 }
 
 static void
 on_ssrc_validated (RTPSession * sess, RTPSource * source)
 {
+  g_object_ref (source);
   RTP_SESSION_UNLOCK (sess);
   g_signal_emit (sess, rtp_session_signals[SIGNAL_ON_SSRC_VALIDATED], 0,
       source);
   RTP_SESSION_LOCK (sess);
+  g_object_unref (source);
 }
 
 static void
 on_ssrc_active (RTPSession * sess, RTPSource * source)
 {
+  g_object_ref (source);
   RTP_SESSION_UNLOCK (sess);
   g_signal_emit (sess, rtp_session_signals[SIGNAL_ON_SSRC_ACTIVE], 0, source);
   RTP_SESSION_LOCK (sess);
+  g_object_unref (source);
 }
 
 static void
 on_ssrc_sdes (RTPSession * sess, RTPSource * source)
 {
+  g_object_ref (source);
   GST_DEBUG ("SDES changed for SSRC %08x", source->ssrc);
   RTP_SESSION_UNLOCK (sess);
   g_signal_emit (sess, rtp_session_signals[SIGNAL_ON_SSRC_SDES], 0, source);
   RTP_SESSION_LOCK (sess);
+  g_object_unref (source);
 }
 
 static void
 on_bye_ssrc (RTPSession * sess, RTPSource * source)
 {
+  g_object_ref (source);
   RTP_SESSION_UNLOCK (sess);
   g_signal_emit (sess, rtp_session_signals[SIGNAL_ON_BYE_SSRC], 0, source);
   RTP_SESSION_LOCK (sess);
+  g_object_unref (source);
 }
 
 static void
 on_bye_timeout (RTPSession * sess, RTPSource * source)
 {
+  g_object_ref (source);
   RTP_SESSION_UNLOCK (sess);
   g_signal_emit (sess, rtp_session_signals[SIGNAL_ON_BYE_TIMEOUT], 0, source);
   RTP_SESSION_LOCK (sess);
+  g_object_unref (source);
 }
 
 static void
 on_timeout (RTPSession * sess, RTPSource * source)
 {
+  g_object_ref (source);
   RTP_SESSION_UNLOCK (sess);
   g_signal_emit (sess, rtp_session_signals[SIGNAL_ON_TIMEOUT], 0, source);
   RTP_SESSION_LOCK (sess);
+  g_object_unref (source);
+}
+
+static void
+on_sender_timeout (RTPSession * sess, RTPSource * source)
+{
+  g_object_ref (source);
+  RTP_SESSION_UNLOCK (sess);
+  g_signal_emit (sess, rtp_session_signals[SIGNAL_ON_SENDER_TIMEOUT], 0,
+      source);
+  RTP_SESSION_LOCK (sess);
+  g_object_unref (source);
 }
 
 /**
@@ -907,16 +964,14 @@ static gboolean
 check_collision (RTPSession * sess, RTPSource * source,
     RTPArrivalStats * arrival, gboolean rtp)
 {
-  /* If we have not arrival address, we can't do collision checking */
-  if (!arrival->have_address) {
+  /* If we have no arrival address, we can't do collision checking */
+  if (!arrival->have_address)
     return FALSE;
-  }
 
   if (sess->source != source) {
     /* This is not our local source, but lets check if two remote
      * source collide
      */
-
     if (rtp) {
       if (source->have_rtp_from) {
         if (gst_netaddress_equal (&source->rtp_from, &arrival->address))
@@ -938,8 +993,9 @@ check_collision (RTPSession * sess, RTPSource * source,
         return FALSE;
       }
     }
-
-    /* In this case, we have third-party collision or loop */
+    /* We received RTP or RTCP from this source before but the network address
+     * changed. In this case, we have third-party collision or loop */
+    GST_DEBUG ("we have a third-party collision or loop");
 
     /* FIXME: Log 3rd party collision somehow
      * Maybe should be done in upper layer, only the SDES can tell us
@@ -1026,7 +1082,7 @@ obtain_source (RTPSession * sess, guint32 ssrc, gboolean * created,
  * rtp_session_get_internal_source:
  * @sess: a #RTPSession
  *
- * Get the internal #RTPSource of @session.
+ * Get the internal #RTPSource of @sess.
  *
  * Returns: The internal #RTPSource. g_object_unref() after usage.
  */
@@ -1042,6 +1098,48 @@ rtp_session_get_internal_source (RTPSession * sess)
   return result;
 }
 
+/**
+ * rtp_session_set_internal_ssrc:
+ * @sess: a #RTPSession
+ * @ssrc: an SSRC
+ *
+ * Set the SSRC of @sess to @ssrc.
+ */
+void
+rtp_session_set_internal_ssrc (RTPSession * sess, guint32 ssrc)
+{
+  RTP_SESSION_LOCK (sess);
+  g_hash_table_steal (sess->ssrcs[sess->mask_idx],
+      GINT_TO_POINTER (sess->source->ssrc));
+
+  sess->source->ssrc = ssrc;
+  rtp_source_reset (sess->source);
+
+  g_hash_table_insert (sess->ssrcs[sess->mask_idx],
+      GINT_TO_POINTER (sess->source->ssrc), sess->source);
+  RTP_SESSION_UNLOCK (sess);
+}
+
+/**
+ * rtp_session_get_internal_ssrc:
+ * @sess: a #RTPSession
+ *
+ * Get the internal SSRC of @sess.
+ *
+ * Returns: The SSRC of the session. 
+ */
+guint32
+rtp_session_get_internal_ssrc (RTPSession * sess)
+{
+  guint32 ssrc;
+
+  RTP_SESSION_LOCK (sess);
+  ssrc = sess->source->ssrc;
+  RTP_SESSION_UNLOCK (sess);
+
+  return ssrc;
+}
+
 /**
  * rtp_session_add_source:
  * @sess: a #RTPSession
@@ -1437,9 +1535,6 @@ rtp_session_process_sr (RTPSession * sess, GstRTCPPacket * packet,
   if (!source)
     return;
 
-  GST_BUFFER_OFFSET (packet->buffer) = source->clock_base;
-  GST_BUFFER_OFFSET_END (packet->buffer) = source->clock_base_time;
-
   prevsender = RTP_SOURCE_IS_SENDER (source);
 
   /* first update the source */
@@ -1669,6 +1764,9 @@ rtp_session_process_rtcp (RTPSession * sess, GstBuffer * buffer,
   if (sess->sent_bye)
     goto ignore;
 
+  /* make writable, we might want to change the buffer */
+  buffer = gst_buffer_make_metadata_writable (buffer);
+
   /* start processing the compound packet */
   more = gst_rtcp_buffer_get_first_packet (buffer, &packet);
   while (more) {
@@ -1752,6 +1850,7 @@ ignore:
  * @buffer: an RTP buffer
  * @current_time: the current system time
  * @ntpnstime: the NTP time in nanoseconds of when this buffer was captured.
+ * This is the buffer timestamp converted to NTP time.
  *
  * Send the RTP buffer in the session manager. This function takes ownership of
  * @buffer.
@@ -2048,6 +2147,7 @@ session_cleanup (const gchar * key, RTPSource * source, ReportData * data)
 {
   gboolean remove = FALSE;
   gboolean byetimeout = FALSE;
+  gboolean sendertimeout = FALSE;
   gboolean is_sender, is_active;
   RTPSession *sess = data->sess;
   GstClockTime interval;
@@ -2090,6 +2190,7 @@ session_cleanup (const gchar * key, RTPSource * source, ReportData * data)
             GST_TIME_ARGS (source->last_rtp_activity));
         source->is_sender = FALSE;
         sess->stats.sender_sources--;
+        sendertimeout = TRUE;
       }
     }
   }
@@ -2105,6 +2206,9 @@ session_cleanup (const gchar * key, RTPSource * source, ReportData * data)
       on_bye_timeout (sess, source);
     else
       on_timeout (sess, source);
+  } else {
+    if (sendertimeout)
+      on_sender_timeout (sess, source);
   }
   return remove;
 }
@@ -2252,6 +2356,7 @@ rtp_session_on_timeout (RTPSession * sess, GstClockTime current_time,
   if (is_rtcp_time (sess, current_time, &data)) {
     if (sess->source->received_bye) {
       /* generate BYE instead */
+      GST_DEBUG ("generating BYE message");
       session_bye (sess, &data);
       sess->sent_bye = TRUE;
     } else {
@@ -2279,6 +2384,7 @@ rtp_session_on_timeout (RTPSession * sess, GstClockTime current_time,
   }
 
   /* check for outdated collisions */
+  GST_DEBUG ("checking collision list");
   item = g_list_first (sess->conflicting_addresses);
   while (item) {
     RTPConflictingAddress *known_conflict = item->data;
@@ -2288,12 +2394,14 @@ rtp_session_on_timeout (RTPSession * sess, GstClockTime current_time,
             RTCP_INTERVAL_COLLISION_TIMEOUT)) {
       sess->conflicting_addresses =
           g_list_delete_link (sess->conflicting_addresses, item);
+      GST_DEBUG ("collision %p timed out", known_conflict);
       g_free (known_conflict);
     }
     item = next_item;
   }
 
   if (sess->change_ssrc) {
+    GST_DEBUG ("need to change our SSRC (%08x)", sess->source->ssrc);
     g_hash_table_steal (sess->ssrcs[sess->mask_idx],
         GINT_TO_POINTER (sess->source->ssrc));
 
@@ -2307,6 +2415,7 @@ rtp_session_on_timeout (RTPSession * sess, GstClockTime current_time,
     sess->bye_reason = NULL;
     sess->sent_bye = FALSE;
     sess->change_ssrc = FALSE;
+    GST_DEBUG ("changed our SSRC to %08x", sess->source->ssrc);
   }
   RTP_SESSION_UNLOCK (sess);
 
@@ -2315,11 +2424,14 @@ rtp_session_on_timeout (RTPSession * sess, GstClockTime current_time,
     /* close the RTCP packet */
     gst_rtcp_buffer_end (data.rtcp);
 
+    GST_DEBUG ("sending packet");
     if (sess->callbacks.send_rtcp)
       result = sess->callbacks.send_rtcp (sess, sess->source, data.rtcp,
-          sess->send_rtcp_user_data);
-    else
+          sess->sent_bye, sess->send_rtcp_user_data);
+    else {
+      GST_DEBUG ("freeing packet");
       gst_buffer_unref (data.rtcp);
+    }
   }
 
   return result;
