@@ -141,6 +141,8 @@ gst_dvbsrc_modulation_get_type (void)
     {QAM_128, "QAM 128", "QAM 128"},
     {QAM_256, "QAM 256", "QAM 256"},
     {QAM_AUTO, "AUTO", "AUTO"},
+    {VSB_8, "8VSB", "8VSB"},
+    {VSB_16, "16VSB", "16VSB"},
     {0, NULL, NULL},
   };
 
@@ -681,6 +683,7 @@ gst_dvbsrc_open_frontend (GstDvbSrc * object)
   char *adapter_desc = NULL;
   gchar *frontend_dev;
   GstStructure *adapter_structure;
+  char *adapter_name = NULL;
 
   frontend_dev = g_strdup_printf ("/dev/dvb/adapter%d/frontend%d",
       object->adapter_number, object->frontend_number);
@@ -715,18 +718,22 @@ gst_dvbsrc_open_frontend (GstDvbSrc * object)
     return FALSE;
   }
 
+  adapter_name = g_strdup (fe_info.name);
+
   object->adapter_type = fe_info.type;
   switch (object->adapter_type) {
     case FE_QPSK:
       adapter_desc = "DVB-S";
       adapter_structure = gst_structure_new ("dvb-adapter",
           "type", G_TYPE_STRING, adapter_desc,
+          "name", G_TYPE_STRING, adapter_name,
           "auto-fec", G_TYPE_BOOLEAN, fe_info.caps & FE_CAN_FEC_AUTO, NULL);
       break;
     case FE_QAM:
       adapter_desc = "DVB-C";
       adapter_structure = gst_structure_new ("dvb-adapter",
           "type", G_TYPE_STRING, adapter_desc,
+          "name", G_TYPE_STRING, adapter_name,
           "auto-inversion", G_TYPE_BOOLEAN,
           fe_info.caps & FE_CAN_INVERSION_AUTO, "auto-qam", G_TYPE_BOOLEAN,
           fe_info.caps & FE_CAN_QAM_AUTO, "auto-fec", G_TYPE_BOOLEAN,
@@ -736,6 +743,7 @@ gst_dvbsrc_open_frontend (GstDvbSrc * object)
       adapter_desc = "DVB-T";
       adapter_structure = gst_structure_new ("dvb-adapter",
           "type", G_TYPE_STRING, adapter_desc,
+          "name", G_TYPE_STRING, adapter_name,
           "auto-inversion", G_TYPE_BOOLEAN,
           fe_info.caps & FE_CAN_INVERSION_AUTO, "auto-qam", G_TYPE_BOOLEAN,
           fe_info.caps & FE_CAN_QAM_AUTO, "auto-transmission-mode",
@@ -745,16 +753,23 @@ gst_dvbsrc_open_frontend (GstDvbSrc * object)
           G_TYPE_BOOLEAN, fe_info.caps % FE_CAN_HIERARCHY_AUTO, "auto-fec",
           G_TYPE_BOOLEAN, fe_info.caps & FE_CAN_FEC_AUTO, NULL);
       break;
+    case FE_ATSC:
+      adapter_desc = "ATSC";
+      adapter_structure = gst_structure_new ("dvb-adapter",
+          "type", G_TYPE_STRING, adapter_desc,
+          "name", G_TYPE_STRING, adapter_name, NULL);
+      break;
     default:
       g_error ("Unknown frontend type: %d", object->adapter_type);
       adapter_structure = gst_structure_new ("dvb-adapter",
           "type", G_TYPE_STRING, "unknown", NULL);
   }
 
-  GST_INFO_OBJECT (object, "DVB card: %s ", fe_info.name);
+  GST_INFO_OBJECT (object, "DVB card: %s ", adapter_name);
   gst_element_post_message (GST_ELEMENT_CAST (object), gst_message_new_element
       (GST_OBJECT (object), adapter_structure));
   g_free (frontend_dev);
+  g_free (adapter_name);
   return TRUE;
 }
 
@@ -834,6 +849,7 @@ gst_dvbsrc_plugin_init (GstPlugin * plugin)
   GST_DEBUG ("binding text domain %s to locale dir %s", GETTEXT_PACKAGE,
       LOCALEDIR);
   bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
+  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
 #endif /* ENABLE_NLS */
 
   return gst_element_register (plugin, "dvbsrc", GST_RANK_NONE,
@@ -1275,6 +1291,11 @@ gst_dvbsrc_tune (GstDvbSrc * object)
         feparams.u.qam.modulation = object->modulation;
         feparams.u.qam.symbol_rate = sym_rate;
         break;
+      case FE_ATSC:
+        GST_INFO_OBJECT (object, "Tuning ATSC to %d", freq);
+        feparams.frequency = freq;
+        feparams.u.vsb.modulation = object->modulation;
+        break;
       default:
         g_error ("Unknown frontend type: %d", object->adapter_type);
 
